-------------------------------------------------------------------------------
--
-- Copyright 2022
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------
-- Author: Job van Wee
-- Purpose: Resize the input data vector so that the output data vector can be
--  stored into the ddr memory.
--
-- Description:
--  The input data gets resized and put into the output data vector.
--
-- Remark:
--  Use VHDL coding template from:
--  https://support.astron.nl/confluence/display/SBe/VHDL+design+patterns+for+RTL+coding
--  The output vector must be larger than the input vector.

LIBRARY IEEE, dp_lib, tech_ddr_lib;
USE IEEE.std_logic_1164.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;

ENTITY ddrctrl_repack IS
  GENERIC (
    g_tech_ddr              : t_c_tech_ddr;                                                                                             -- type of memory
    g_in_data_w	            : NATURAL                               := 168                                                              -- the input data with
  );
  PORT (
    clk	     	              : IN  STD_LOGIC;
    in_data  	              : IN  STD_LOGIC_VECTOR(g_in_data_w-1 DOWNTO 0);
    out_of      	          : OUT NATURAL                           := 0;
    out_sosi                : OUT t_dp_sosi                         := c_dp_sosi_init
  );
END ddrctrl_repack;


ARCHITECTURE rtl OF ddrctrl_repack IS

  CONSTANT c_out_data_w     : NATURAL                               := func_tech_ddr_ctlr_data_w( g_tech_ddr );                         -- the output data with, 576
  CONSTANT k_c_v_w			    : NATURAL                               := c_out_data_w*2;                                                  -- the c_v data with, 2*576=1152

  SIGNAL  c_v_count			    : NATURAL                               := 0;                                                               -- the amount of times the c_v vector received data from the input since the last time it was filled completely
  SIGNAL  out_data_count		: NATURAL                               := 0;                                                               -- the amount of times the output data vector has been filled since the last time c_v was filled completely

BEGIN

  p_clk : PROCESS(clk)

  VARIABLE a_of             : NATURAL                               := 0;                                                               -- amount of overflow
  VARIABLE c_v				      : STD_LOGIC_VECTOR (k_c_v_w-1 DOWNTO 0) := (OTHERS => '0');                                                 -- the vector that stores the input data until the data is put into the output data vector

  BEGIN

    IF rising_edge(clk) THEN

      IF ((g_in_data_w*(c_v_count+1))+a_of >= c_out_data_w*(out_data_count+1)) THEN								                                      -- if the input data exceeds the output data vector width

	      IF (out_data_count = 1) THEN													                                                                         	-- if the input data exceeds c_v widt
          a_of := a_of + (g_in_data_w*(c_v_count+1)) - (c_out_data_w*(out_data_count+1));							                                  -- check how much overflow there is and safe it in a_of
        	out_of <= a_of;													                                                                                 		  -- set the output overflow to the overflow that maches the out_sosi.data vector
	        c_v(k_c_v_w - 1 DOWNTO k_c_v_w-(g_in_data_w - a_of)) := in_data(g_in_data_w - a_of - 1 DOWNTO 0);			                        -- fill the rest of c_v untill the end
	        c_v(a_of - 1 DOWNTO 0) := in_data(g_in_data_w - 1 DOWNTO g_in_data_w - a_of);							                                    -- fill the start of c_v untill the a_of
	        out_sosi.data(c_out_data_w - 1 DOWNTO 0) <= c_v(k_c_v_w - 1 DOWNTO c_out_data_w);							                                -- fill out_sosi.data with 2nd part of c_v
	        out_sosi.valid <= '1';													                                                          	                  -- out_sosi.valid 1
	        c_v_count	<= 0; 													                                                                                   	-- reset counter
          out_data_count <= 0;													                                                                               	-- reset counter

	      Else															                                                                                            	-- if the input data exceeds output data vector width but not the c_v vector widt
	        c_v(g_in_data_w * (c_v_count + 1) + a_of - 1 DOWNTO g_in_data_w * c_v_count + a_of) := in_data(g_in_data_w - 1 DOWNTO 0); 	  -- fill c_v
	        c_v_count <= c_v_count + 1;													                                                                          -- increase the counter of c_v with 1
	        out_sosi.data(c_out_data_w - 1 DOWNTO 0) <= c_v(c_out_data_w - 1 DOWNTO 0);						                                        -- fill out_sosi.data with 1st part of c_v
	        out_sosi.valid <= '1';													                                                                              -- out_sosi.valid 1
	        out_data_count <= out_data_count + 1;													                                                                -- increase the counter of out_sosi.data with 1
	      END IF;

      ELSE															                                                                                              	-- if the input data doesn't exceeds the output data vector width
        c_v(g_in_data_w * (c_v_count + 1) + a_of - 1 DOWNTO g_in_data_w * c_v_count + a_of) := in_data(g_in_data_w - 1 DOWNTO 0);	      -- fill c_v
	      c_v_count <= c_v_count + 1;														                                                                        	-- increase the counter of c_v with 1
        out_sosi.valid <= '0';														                                                                              -- out_sosi.valid 0

      END IF;

    END IF;

  END PROCESS;

END rtl;


